/* A simple VWINL file browser -- Williams */

#include "vwinl.h"
#include "browse.h"
#include <stdio.h>
#include <string.h>
#include <commdlg.h>

#define TOPMARGIN 5
#define LEFTMARGIN 5


/* Filters for common dialog */
char filefilter[]="All files (*.*)\0*.*\0"
                  "Text files (*.txt)\0*.c\0"
                  "C files (*.c)\0*.c\0"
                  "H files (*.h)\0*.h\0"
                  "C++ files (*.cpp)\0*.cpp\0"
                  "DEF files (*.def)\0*.def\0"
                  "RC files (*.rc)\0*.rc\0\0";

/* Scroll sizes */
int xscroll,yscroll,xpage,ypage;


/* This routine computes the text extent and properly
   modifies the wid and hi variables for the caller.
   You could use GetTextExtent() inline for this
   except that WIN32 requires a slightly differnt
   approach */
void text_extent(HDC dc,char *s,UINT ct,UINT *wid,UINT *hi)
  {
#ifdef WIN32
  SIZE extent;
  GetTextExtentPoint(dc,s,ct,&extent);
  if (wid) *wid=max(*wid,(UINT)extent.cx);
  if (hi) *hi+=extent.cy;
#else
  DWORD extent;
  extent=GetTextExtent(dc,s,ct);
  if (wid) *wid=max(*wid,LOWORD(extent));
  if (hi) *hi+=HIWORD(extent);
#endif
  }

/* Open file -- this does *ALL* the work! */
void open_file(HWND w)
  {
  OPENFILENAME ofile;
  FILE *in;
  VMAP *map;
  HCURSOR cursor;
  char fn[256],ft[256];
  int err;
  unsigned wid=0,hi=0,y=TOPMARGIN;
  unsigned scr_wid=0,scr_hi=0;
  HDC dc;
/* Use common dialog to get file name */
  memset(&ofile,0,sizeof(OPENFILENAME));
  *fn=*ft='\0';
  ofile.lStructSize=sizeof(OPENFILENAME);
  ofile.hwndOwner=w;
  ofile.lpstrFilter=filefilter;
  ofile.nFilterIndex=1;
  ofile.lpstrFile=fn;
  ofile.nMaxFile=sizeof(fn);
  ofile.lpstrFileTitle=ft;
  ofile.nMaxFileTitle=sizeof(ft);
  ofile.Flags=OFN_HIDEREADONLY|OFN_PATHMUSTEXIST|
              OFN_FILEMUSTEXIST;
  if (!(err=GetOpenFileName(&ofile))||!(in=fopen(fn,"r")))
    {
/* If err is FALSE then might just be a cancel
    CommDlgExtendedError returns 0 if it was a cancel */
    if (err||CommDlgExtendedError())
       MessageBox(w,"File open error",NULL,
                  MB_ICONSTOP|MB_OK);
    return;
    }
/* Set window title */
  wsprintf(fn,"BROWSE - %s",(char far *)ft);
  SendMessage(w,WM_SETTEXT,0,(LONG)fn);
/* Wait cursor */
  cursor=SetCursor(LoadCursor(NULL,IDC_WAIT));
/* Make window have no map */
  map=Vselect_map(w,NULL);
  if (map)  /* If there was a map there, kill it */
    Vdestroy_map(map);
  dc=GetDC(w);
/* Read file to compute width and length of VMAP */
  while (fgets(ft,sizeof(ft),in))
    {
    int l;
    ft[l=(strlen(ft)-1)]='\0';
/* Make empty lines have one blank */
    if (!l)
      {
      ft[0]=' ';
      ft[1]='\0';
      l=1;
      }
/* compute extents */
    text_extent(dc,ft,l,&wid,&hi);
    }
  ReleaseDC(w,dc);
/* create map of the right size */
  map=Vcreate_map(wid+2*LEFTMARGIN,hi+2*TOPMARGIN);
  if (!map)
    {
    MessageBox(w,"File too large!",NULL,MB_ICONSTOP|MB_OK);
    SendMessage(w,WM_SETTEXT,0,(LONG)"BROWSE");
    SetCursor(cursor);
    return;
    }
/* Start file over */
  rewind(in);
/* Select map into window */
  Vselect_map(w,map);
  dc=Vget_vdc(w);
/* Set scroll dimensions */
  text_extent(dc," ",1,&scr_wid,&scr_hi);
  Vset_scroll(map,xscroll=scr_wid,yscroll=scr_hi,
             xpage-scr_wid,ypage-scr_hi);
/* Read file into VMAP */
  while (fgets(ft,sizeof(ft),in))
    {
    int l;
    ft[l=(strlen(ft)-1)]='\0';
/* Empty lines = 1 space */
    if (!l)
      {
      ft[0]=' ';
      ft[1]='\0';
      l=1;
      }
    TextOut(dc,LEFTMARGIN,y,ft,l);
    text_extent(dc,ft,l,NULL,&y);
    }
/* Done with file */
  fclose(in);
/* Restore cursor */
  SetCursor(cursor);
/* Force drawing update */
  Vcommit_draw(w);
  }

/* Do menu functions */
void do_menu(HWND w,UINT wParam)
   {
/* LOWORD for WIN32 compatibility */
   switch (LOWORD(wParam))
      {
      case MENU_ABOUT:
        MessageBox(w,
          "Browse -- A VWINL File Browser by Al Williams",
          "About BROWSE",MB_OK);
          break;

       case MENU_EXIT:
          DestroyWindow(w);
          break;

       case MENU_OPEN:
          open_file(w);
          break;
         }
      }

/* Our callback */
long usr_cb(HWND w, UINT Message,
            UINT wParam, LONG lParam)
  {
  switch (Message)
     {
     VMAP *map;
/* Handle menu commands */
     case WM_COMMAND:
       do_menu(w,wParam);
       break;

/* If window resizes, recompute size of a scrolling
   page */
    case WM_SIZE:
       xpage=LOWORD(lParam);
       ypage=HIWORD(lParam);
       map=Vget_map(w);
       if (map)
         Vset_scroll(map,xscroll,yscroll,
                     xpage-xscroll,ypage-yscroll);
       break;

     default:
       return DefWindowProc(w,Message,wParam,lParam);
     }
  return 0;
  }

/* VWINL main function */
int main(HANDLE hInstance, HANDLE hPrevInstance,
                   LPSTR lpszCmdLine, int nCmdShow )
   {
   HWND w;
   Vset_monomode(TRUE);
   if (Vcreate_window("BROWSE",
                  WS_OVERLAPPEDWINDOW,CW_USEDEFAULT,0,
                  CW_USEDEFAULT,0,NULL,"mainmenu",
                  usr_cb,V_AUTOHSCROLL|V_AUTOVSCROLL|
                  V_NOMAP|V_KSCROLL,NULL,&w,nCmdShow))
       {
       MessageBox(NULL,"Can't create window",NULL,MB_OK);
       return 0;
       }
   return 1;
   }
