/* Simple-minded resource display using VWINL */

#ifdef WIN32
#error FREESHOW needs TOOLHELP and will not work with WIN32
#endif

#include "vwinl.h"
#include <toolhelp.h>
#include "freeshow.h"

/* Default window width/height */
#define WIDTH 160
#define HEIGHT 150
/* Timer interval in ms (1000=1 second) */
#define TIME_SLICE 1000

HANDLE our_instance;

/* These flags determine if number and/or histogram show */
int show_num=1;
int show_hist=1;

/* Calculate free GDI/USER heap
   See Undocumented Windows by Schulman, Maxey, & Pietrek
   to learn more about how this works */
void get_free_info(HWND w)
  {
  char str[5];
  SYSHEAPINFO info;
  HDC dc=Vget_vdc(w);
  WORD percent;
  HBRUSH brush;
  info.dwSize=sizeof(info);
  SystemHeapInfo(&info);
/* Which one is smallest? */
  percent=min(info.wUserFreePercent,info.wGDIFreePercent);
/* Clear whole VMAP */
  Vclear_win(w);
/* Show number if asked */
  if (show_num)
    {
    wsprintf(str,"%3d%%",percent);
    TextOut(dc,10,10,str,4);
    }
/* Show histogram if asked */
  if (show_hist)
    {
    COLORREF color;
/* Draw 100% rectangle */
    Rectangle(dc,10,40,110,50);
/* Pick red, green or yellow depending on range */
     if (percent<40)
       color=RGB(0xFF,0,0);
     else if (percent>60)
       color=RGB(0,0xFF,0);
     else
       color=RGB(0xFF,0xFF,0);  /* yellow */
    brush=CreateSolidBrush(color);
    brush=SelectObject(dc,brush);
/* Draw histogram */
    Rectangle(dc,10,40,10+percent,50);
/* Get rid of brush */
    DeleteObject(SelectObject(dc,brush));
    }
/* Always print title */
  TextOut(dc,10,60,"% Free Resources",16);
/* Commit it! */
  Vcommit_draw(w);
  }


/* Do menu functions */
void do_menu(HWND w,WORD wParam)
   {
   switch (wParam)
      {
      HMENU menu;
      case MENU_ABOUT:
        MessageBox(w,
                  "System Resource Display by Al Williams",
                  "About FREESHOW",MB_OK);
         break;

         case MENU_EXIT:
              DestroyWindow(w);
              break;

/* Toggle numeric display */
         case MENU_NUM:
              menu=GetMenu(w);
              show_num^=1;
              CheckMenuItem(menu,MENU_NUM,
                   show_num?MF_CHECKED:MF_UNCHECKED);
              break;

/* Toggle histogram display */
         case MENU_HIST:
              menu=GetMenu(w);
              show_hist^=1;
              CheckMenuItem(menu,MENU_HIST,
                   show_hist?MF_CHECKED:MF_UNCHECKED);
              break;

/* Toggle scaling */
        case MENU_SCALE:
             menu=GetMenu(w);
             if (Vset_flags(w,V_SCALE|V_AUTOHSCROLL|V_AUTOVSCROLL,VF_TOG)
                 &V_AUTOHSCROLL)
               CheckMenuItem(menu,MENU_SCALE,MF_CHECKED);
             else
               CheckMenuItem(menu,MENU_SCALE,MF_UNCHECKED);
             break;

/* Reset window size */
         case MENU_RESET:
             {
             RECT r;
/* Make sure window isn't maxxed out -- if it is bizzare things
   will happen! */
             ShowWindow(w,SW_SHOWNORMAL);
             GetWindowRect(w,&r);
             MoveWindow(w,r.left,r.top,WIDTH,HEIGHT,TRUE);
             break;
             }
         }
     }


/* Our callback */
long usr_cb(HWND w, UINT Message,
            UINT wParam, LONG lParam)
  {
  switch (Message)
     {
     HDC dc;
/* Start timer */
     case WM_VCREATE:
       SetTimer(w,1,TIME_SLICE,NULL);
       break;

/* menu processing */
     case WM_COMMAND:
       do_menu(w,wParam);
       break;


/* Do work when the clock chimes */
     case WM_TIMER:
       get_free_info(w);
       break;

/* On exit, release timer */
     case WM_CLOSE:
       KillTimer(w,1);
       DestroyWindow(w);
       break;

     default:
       return DefWindowProc(w,Message,wParam,lParam);
     }
  return 0;
  }

/* VWINL main function */
int main(HANDLE hInstance, HANDLE hPrevInstance,
                   LPSTR lpszCmdLine, int nCmdShow )
   {
   HWND w;
   our_instance=hInstance;
   if (Vcreate_window("Free Show",
                        WS_OVERLAPPEDWINDOW,CW_USEDEFAULT,0,
                        WIDTH,HEIGHT,NULL,"mainmenu",
                        usr_cb,V_AUTOHSCROLL|V_AUTOVSCROLL,
                        NULL,&w,nCmdShow))
       {
       MessageBox(NULL,"Can't create window",NULL,MB_OK);
       return 0;
       }
   return 1;
   }
