/* VWIN virtual window package -- Al Williams */
#include "vwinl.h"
#include <windowsx.h>
#include <string.h>

/* Local prototypes */
static void do_paint(HWND);
long WINAPI VWndProc (HWND w, UINT Message,
                          UINT wParam, LONG lParam);
static void save_info(HWND w,MEMWINFO *info);
static void set_sb(HWND w,MEMWINFO *minfo,UINT wid,
                   UINT hi,int save);
static void scrollit(HWND w,MEMWINFO *minfo,int type,
                     WORD code,HWND sb,WORD pos);
static void key_scroll(HWND w,UINT key);


/* Global variables */
static HANDLE Hinstance;  /* Our instance */
static int monomode;      /* Make mono bitmaps? */
/* Flag to tell us if user allows us to quit */
static int V_quit=0;


/* VWINL WinMain -- this calls our main() function.
   If main() returns 0, then we abort. */

int PASCAL WinMain(HANDLE hInstance, HANDLE hPrevInstance,
                   LPSTR lpszCmdLine, int nCmdShow )
{
    WNDCLASS    wndClass;
    MSG         msg;
    HACCEL      haccel;

    /*
     * Register window class style if first
     * instance of this program.
     */
    Hinstance=hInstance;
    if ( !hPrevInstance )
    {
/* NOTE: VWINL assumes the window will have a common DC!
   Don't use CS_OWNDC or CS_PARENTDC unless you know
   what you are getting into! */
        wndClass.style=CS_HREDRAW|CS_VREDRAW|CS_DBLCLKS;
        wndClass.lpfnWndProc=(WNDPROC)VWndProc;
        wndClass.cbClsExtra=0;
        wndClass.cbWndExtra=
           sizeof(MEMWINFO)+(sizeof(MEMWINFO)%2);
        wndClass.hInstance=hInstance;
        wndClass.hIcon=LoadIcon(hInstance,"vappicon");
        wndClass.hCursor=LoadCursor(NULL, IDC_ARROW );
        wndClass.hbrBackground=GetStockObject(WHITE_BRUSH);
        wndClass.lpszMenuName=NULL;
        wndClass.lpszClassName="VWINL";
    if (!RegisterClass(&wndClass))
        return FALSE;
    }

    if (!main(hInstance,hPrevInstance,
                   lpszCmdLine, nCmdShow )) return FALSE;

/* Try to load an accelerator -- no big deal if it
   isn't there */
      haccel=LoadAccelerators(hInstance,"VACCEL");

/* Sorta ordinary message loop -- will translate accelerators
   only if appropriate */
      while (GetMessage(&msg, NULL, 0, 0))
       {
       if (!haccel||
         !TranslateAccelerator(msg.hwnd, haccel, &msg))
         {
         TranslateMessage(&msg);
         DispatchMessage(&msg);
         }
      }
      return 0;
   }



/* Main Window procedure */
long APIENTRY VWndProc (HWND w, UINT Message,
                          UINT wParam, LONG lParam)
    {
    MEMWINFO minfo;
    Vget_info(w,&minfo);
    switch(Message)
    {
/* Don't let user see WM_CREATE -- we aren't ready for him
   yet */
        case WM_CREATE:
            return DefWindowProc(w,Message,wParam,lParam);

/* Always do painting and don't tell user */
        case WM_PAINT:
            do_paint(w);
            return 0;

        case WM_SIZE:
/* Catch 1st size */
            if (minfo.flags&V_INIT)
              {
              if ((LOWORD(lParam)!=minfo.width||
                  HIWORD(lParam)!=minfo.height)&&minfo.map)
                {
                Vresize_winmap(w,LOWORD(lParam),HIWORD(lParam));
                }
              minfo.flags&=~V_INIT;
              save_info(w,&minfo);
              break;
              }
/* Set scrollbars */
            if (minfo.map)
              {
/* Everytime you toggle a scroll bar from on to off or off to on
   you get a WM_SIZE message! This little state machine lets
   us turn off the bars without getting into an endless loop. */

              static sizelock=0;
              RECT r;
              if (sizelock==1) return 0;
              if (sizelock!=2)
                {
/* Turn off both scroll bars so set_sb() can use the whole
   client area if that's what it needs */
                sizelock=1;
                if (minfo.flags&V_AUTOVSCROLL)
                  SetScrollRange(w,SB_VERT,0,0,FALSE);
                if (minfo.flags&V_AUTOHSCROLL)
                  SetScrollRange(w,SB_HORZ,0,0,FALSE);
/* Size might have changed, so reset it */
                GetClientRect(w,&r);
                lParam=MAKELONG(r.right-r.left,r.bottom-r.top);
                }
/* Turn scroll bars on or off */
              sizelock=2;
              set_sb(w,&minfo,LOWORD(lParam),HIWORD(lParam),TRUE);
/* Size might have changed again, so reset it */
              GetClientRect(w,&r);
              lParam=MAKELONG(r.right-r.left,r.bottom-r.top);
              sizelock=0;
              }
/* Handle V_RESIZE if active */
            if (minfo.flags&V_RESIZE&&minfo.map)
              {
              Vresize_winmap(w,LOWORD(lParam),HIWORD(lParam));
              }
            break;

/* Scroll cases */
        case WM_VSCROLL:
            if (minfo.flags&V_AUTOVSCROLL)
              scrollit(w,&minfo,SB_VERT,
              GET_WM_VSCROLL_CODE(wParam,lParam),
              GET_WM_VSCROLL_HWND(wParam,lParam),
              GET_WM_VSCROLL_POS(wParam,lParam));
            break;

        case WM_HSCROLL:
           if (minfo.flags&V_AUTOHSCROLL)
              scrollit(w,&minfo,SB_HORZ,
              GET_WM_HSCROLL_CODE(wParam,lParam),
              GET_WM_HSCROLL_HWND(wParam,lParam),
              GET_WM_HSCROLL_POS(wParam,lParam));
           break;

        case WM_KEYDOWN:
            if (minfo.flags&V_KSCROLL) key_scroll(w,wParam);
            break;

        case WM_DESTROY:
/* pass to user if V_NOQUIT set */
            if (minfo.flags&V_NOQUIT) break;
            V_quit=1;
/* Pass to user, if V_quit is set, go ahead and kill ourselves */
            if (!minfo.cb(w,Message,wParam,lParam)&&V_quit)
              {
              /* Clean up window's resources here... */
              if (minfo.map&&minfo.map->refct&&!(minfo.flags&V_NOFREEMAP))
                {
                minfo.map->refct--;
                Vdestroy_map(minfo.map);
                }
              PostQuitMessage(0);
              return 0;
              }
            return 0;
       }
/* pass to user's callback */
      if (minfo.cb)
        return minfo.cb(w,Message,wParam,lParam);
      else
        return DefWindowProc(w,Message,wParam,lParam);

    }


/* Create a VWIN -- see text for description
   Returns 0 for success */
int Vcreate_window(char *title,DWORD style,int x,int y,
      int wid,int hi,HWND parent,LPCSTR menu,
      long (*cb)(HWND,UINT,UINT,LONG),
      unsigned long vflags,HDC *dc,HWND *win,int show)
    {
    HWND w;
    HMENU hMenu=NULL;
    RECT r;
    MEMWINFO minfo;
    if (menu&&!parent)
      hMenu=LoadMenu(Hinstance,menu);
    else
      hMenu=(HMENU)menu;
/* Auto set scroll style */
    if (vflags&V_AUTOHSCROLL) style|=WS_HSCROLL;
    if (vflags&V_AUTOVSCROLL) style|=WS_VSCROLL;
    memset(&minfo,0,sizeof(MEMWINFO));
    minfo.flags=vflags|V_INIT;
    minfo.cb=cb;
    w=*win=CreateWindow("VWINL",title,style,x,y,
                  wid,hi,parent,hMenu,Hinstance,NULL);
    if (!*win) return 1;
    save_info(*win,&minfo);

    SetScrollRange(*win,SB_HORZ,0,0,TRUE);
    SetScrollRange(*win,SB_VERT,0,0,TRUE);


    GetClientRect(w,&r);

   /* create DC the same size */
    minfo.xoff=minfo.yoff=0;
    minfo.width=r.right-r.left;
    minfo.height=r.bottom-r.top;
    if ((minfo.flags&V_NOMAP)==0)
      {
      minfo.map=Vcreate_map(minfo.width,minfo.height);
      if (!minfo.map) return 2;
      if (dc) *dc=minfo.map->dc;
      }
    else
      minfo.map=NULL;
  /* Store cb and other data in extra words */
    save_info(w,&minfo);

/* finish up */
    ShowWindow(*win,show);
/* Call user's callback with WM_VCREATE */
    minfo.cb(*win,WM_VCREATE,0,0);
    UpdateWindow(*win);
    return 0;
    }


/* Make a VMAP -- respects monomode flag */
VMAP *Vcreate_map(int wid,int hi)
  {
  HDC dc;
  VMAP *bm;
  HWND w;
  w=GetDesktopWindow();  /* any window will do */
  dc=GetDC(w);
  if (!dc) return NULL;
/* temp use of malloc */
  bm=(VMAP *)LocalAlloc(LPTR,sizeof(VMAP));
  if (!bm) return NULL;
  bm->dc=CreateCompatibleDC(dc);
/* release Desktop DC */
  ReleaseDC(w,dc);
  if (!bm->dc)
    {
    LocalFree((HLOCAL)bm);
    return NULL;
    }
/* Make the bitmap */
/* NOTE: Windows won't let you make an arbitrary colored
   bitmap. You must have a device that corresponds to
   the color-size of the bitmap */
  if (!monomode)
    bm->bitmap=CreateBitmap(wid,hi,GetDeviceCaps(dc,PLANES),
                        GetDeviceCaps(dc,BITSPIXEL),NULL);
  else
    bm->bitmap=CreateBitmap(wid,hi,1,1,NULL); /* mono */

  if (!bm->bitmap)
    {
    DeleteDC(bm->dc);
    LocalFree((HLOCAL)bm);
    return NULL;
    }
/* Note: This is supposed to be in .1 mm units, but since no
   one else uses it, who cares! */
#ifndef WIN32
  SetBitmapDimension(bm->bitmap,wid,hi);
#else
  SetBitmapDimensionEx(bm->bitmap,wid,hi,NULL);
#endif
  bm->defbitmap=SelectObject(bm->dc,bm->bitmap);
  if (!bm->defbitmap)
    {
    DeleteDC(bm->dc);
    LocalFree((HLOCAL)bm);
    return NULL;
    }
/* Set default stuff */
  bm->xstep=bm->ystep=1;
  bm->xpage=bm->ypage=10;
  bm->refct=1;
  bm->stretch_mode=BLACKONWHITE;
  Vclear_map(bm);
  return bm;
  }

/* Free up a valid VMAP if its refct is 1 or 0 */
void Vdestroy_map(VMAP *map)
  {
  if (map->refct>1)
    return;
  SelectObject(map->dc,map->defbitmap);
  DeleteObject(map->bitmap);
  DeleteDC(map->dc);
  LocalFree((HLOCAL)map);
  }

/* Associate a new map (or NULL for no map)
   with a window -- returns the old map */
VMAP *Vselect_map(HWND w,VMAP *new)
  {
  MEMWINFO minfo;
  VMAP *rc;
#ifndef WIN32
  DWORD dims;
#else
  SIZE dims;
#endif
  RECT r;
  GetClientRect(w,&r);
  Vget_info(w,&minfo);
  rc=minfo.map;
  minfo.map=new;
  if (rc) rc->refct--;
/* NULL is OK here in which case we don't do much */
  if (new)
    {
    new->refct++;
#ifndef WIN32
    dims=GetBitmapDimension(minfo.map->bitmap);
    minfo.width=LOWORD(dims);
    minfo.height=HIWORD(dims);
#else
    GetBitmapDimensionEx(minfo.map->bitmap,&dims);
    minfo.width=dims.cx;
    minfo.height=dims.cy;
#endif
    }
  else
    {
    minfo.width=minfo.height=0;
    }
  save_info(w,&minfo);
  set_sb(w,&minfo,r.right-r.left,r.bottom-r.top,TRUE);
  Vcommit_draw(w);
  return rc;
  }


/* Set stretch mode for map -- returns old mode */
int Vset_stretchmode(VMAP *m,int mode)
  {
  int rv;
  rv=m->stretch_mode;
  m->stretch_mode=mode;
  return rv;
  }

/* Sets global monomode flag which causes VWINL to create
   monochrome bitmaps to save space.
   Returns old value (of course) */
int Vset_monomode(int mode)
  {
  int rv;
  rv=monomode;
  monomode=mode;
  return rv;
  }


/* Get window's map */
VMAP *Vget_map(HWND w)
  {
  MEMWINFO minfo;
  Vget_info(w,&minfo);
  return minfo.map;
  }

/* Erase a map's surface */
void Vclear_map(VMAP *m)
  {
  HBRUSH brush;
  unsigned int wid,hi;
#ifndef WIN32
  DWORD dims;
#else
  SIZE dims;
#endif
/* We store bitmap dimension this way.
   Units are pixels contrary to the .1mm convention */
#ifndef WIN32
  dims=GetBitmapDimension(m->bitmap);
  wid=LOWORD(dims);
  hi=HIWORD(dims);
#else
  GetBitmapDimensionEx(m->bitmap,&dims);
  wid=dims.cx;
  hi=dims.cy;
#endif
/* make background brush */
  brush=CreateSolidBrush(GetBkColor(m->dc));
  brush=SelectObject(m->dc,brush);
/* Brush area */
  PatBlt(m->dc,0,0,wid,hi,PATCOPY);
  DeleteObject(SelectObject(m->dc,brush));
  }


/* Change a window's VWINL flags -- this only
   makes sense for some flags. For example, V_NOMAP is
   meaningless here. If you ever plan to set V_AUTOHSCROLL
   or V_AUTOVSCROLL, make sure to set the scroll bar style
   flags during Vcreate_window (this happens automatically
   when you set V_AUTOxSCROLL during the create.

   Returns old flag value */
unsigned long Vset_flags(HWND w,unsigned long flags,int cmd)
  {
  unsigned long  rv;
  MEMWINFO minfo;
  RECT r;
  Vget_info(w,&minfo);
  GetClientRect(w,&r);
  rv=minfo.flags;
  switch (cmd)
    {
    case VF_SET:
       minfo.flags|=flags;
       break;
    case VF_CLR:
       minfo.flags&=~flags;
       break;
    case VF_TOG:
       minfo.flags^=flags;
       break;
    default:
       minfo.flags=flags;
       break;
    }
  save_info(w,&minfo);
  if (((rv&V_AUTOHSCROLL)^(minfo.flags&V_AUTOHSCROLL))
      ||((rv&V_AUTOVSCROLL)^(minfo.flags&V_AUTOVSCROLL)))
    {
/* scroll changed */
    if (!(minfo.flags&V_AUTOHSCROLL))
      {
      minfo.xoff=0;
      SetScrollRange(w,SB_HORZ,0,0,TRUE);
      }
    if (!(minfo.flags&V_AUTOVSCROLL))
      {
      minfo.yoff=0;
      SetScrollRange(w,SB_VERT,0,0,TRUE);
      }
    save_info(w,&minfo);
    if (minfo.flags&(V_AUTOHSCROLL|V_AUTOVSCROLL))
      set_sb(w,&minfo,r.right-r.left,r.bottom-r.top,TRUE);
    Vcommit_draw(w);
    }
  return rv;
  }

/* Set the VMAP offset in pixels */
void Vset_offset(HWND w,int x,int y)
  {
  MEMWINFO minfo;
  Vget_info(w,&minfo);
  minfo.xoff=x;
  minfo.yoff=y;
  save_info(w,&minfo);
  }

/* Read the VMAP pixel offsets */
void Vget_offset(HWND w,int *x,int *y)
  {
  MEMWINFO minfo;
  Vget_info(w,&minfo);
  if (x) *x=minfo.xoff;
  if (y) *y=minfo.yoff;
  }


/* Get window's VMAP dc */
HDC Vget_vdc(HWND w)
  {
  MEMWINFO minfo;
  Vget_info(w,&minfo);
  return minfo.map->dc;
  }


/* Resize a map
   Returns 0 if OK */
int Vresize_map(VMAP *m,int wid,int hi)
  {
  VMAP *newmap;
  int oldstate;
  HBITMAP oldbm=m->bitmap;
  newmap=Vcreate_map(wid,hi);
  if (!newmap) return 1;
  oldstate=SetMapMode(m->dc,MM_TEXT);
  BitBlt(newmap->dc,0,0,wid,hi,m->dc,0,0,SRCCOPY);
/* copy the right parts from newmap */
  m->bitmap=newmap->bitmap;
  SetMapMode(m->dc,oldstate);
/* Except for bitmap! */
  SelectObject(newmap->dc,newmap->defbitmap);
  SelectObject(m->dc,m->bitmap);
/* Delete old dc and bitmap */
  DeleteDC(newmap->dc);
  DeleteObject(oldbm);
  LocalFree((HLOCAL)newmap);
  return 0;
  }

/* Resize VMAP attached to window -- returns 0 if OK */
int Vresize_winmap(HWND w,int wid,int hi)
  {
  MEMWINFO minfo;
  RECT r;
  GetClientRect(w,&r);
  Vget_info(w,&minfo);
  if (Vresize_map(minfo.map,wid,hi)) return 1;
  minfo.width=wid;
  minfo.height=hi;
  save_info(w,&minfo);
  set_sb(w,&minfo,r.right-r.left,r.bottom-r.top,TRUE);
  return 0;
  }



/* Clear quit flag for user */
void Vdont_quit()
  {
  V_quit=0;
  }


/* Get VWIN info -- public */
void Vget_info(HWND w,MEMWINFO *info)
  {
  int i;
  for (i=0;i<sizeof(MEMWINFO);i+=2)
    *(unsigned short *)(((unsigned char *)info)+i)
        =GetWindowWord(w,i);
  }

/* Save VWIN info (local use only) */
static void save_info(HWND w,MEMWINFO *info)
  {
  int i;
  for (i=0;i<sizeof(MEMWINFO);i+=2)
    SetWindowWord(w,i,*(unsigned short *)
                 (((unsigned char *)info)+i));
  }



/* Scroll handler (local) */
static void scrollit(HWND w,MEMWINFO *minfo,int type,
                     WORD code,HWND sb,WORD pos)
  {

  unsigned int *offset;
/* Store as long to avoid unsigned underflow */
  long newoffset;
  int step,page;
  RECT r;
  GetClientRect(w,&r);
/* Set up offset and steps */
  if (type==SB_VERT)
    {
    offset=&minfo->yoff;
    step=minfo->map->ystep;
    page=minfo->map->ypage;
    }
  else
    {
    offset=&minfo->xoff;
    step=minfo->map->xstep;
    page=minfo->map->xpage;
    }
  newoffset=*offset;

/* Process scroll command */
  switch (code)
    {
    case SB_TOP:
       newoffset=0;
       break;

    case SB_BOTTOM:
       if (type==SB_VERT)
         {
         newoffset=minfo->height-(r.bottom-r.top);
         }
       else
         {
         newoffset=minfo->width-(r.right-r.left);
         }
       break;

    case SB_LINEUP:
       step=-step;
/* fall thru */
    case SB_LINEDOWN:
       newoffset+=step;
       break;

    case SB_PAGEUP:
       page=-page;
/* fall thru */
    case SB_PAGEDOWN:
       newoffset+=page;
       break;

    case SB_THUMBPOSITION:
       newoffset=pos;
       break;
/* I didn't process SB_THUMBTRACK since a big hires VMAP can take
   too long to paint */
    }
    if (newoffset<0) newoffset=0;
/* Update the offset */
    if (type==SB_VERT)
      {
      if (newoffset+(r.bottom-r.top)>minfo->height)
         newoffset=minfo->height-(r.bottom-r.top);
      }
    else
      {
      if (newoffset+(r.right-r.left)>minfo->width)
         newoffset=minfo->width-(r.right-r.left);
      }
/* Update position */
    SetScrollPos(w,type,(unsigned)newoffset,TRUE);
    *offset=newoffset;
    save_info(w,minfo);
    Vcommit_draw(w);
    }

/* Set scroll parameters
   Currently you can't read them back unless you call
   Vget_info(). If you must have a Vget_scroll call you
   can write it! */
void Vset_scroll(VMAP *m,int xstep,int ystep,int xpage,int ypage)
  {
  m->xstep=xstep;
  m->ystep=ystep;
  m->xpage=xpage;
  m->ypage=ypage;
  return;
  }

/* Process "scroll" keys */
static void key_scroll(HWND w,UINT key)
  {
  switch (key)
    {
    case VK_HOME:
      SendMessage(w,WM_VSCROLL,SB_TOP,0L);
      break;

    case VK_END:
      SendMessage(w,WM_VSCROLL,SB_BOTTOM,0L);
      break;

    case VK_PRIOR:
      SendMessage(w,WM_VSCROLL,SB_PAGEUP,0L);
      break;

    case VK_NEXT:
      SendMessage(w,WM_VSCROLL,SB_PAGEDOWN,0L);
      break;

    case VK_UP:
      SendMessage(w,WM_VSCROLL,SB_LINEUP,0L);
      break;

    case VK_DOWN:
      SendMessage(w,WM_VSCROLL,SB_LINEDOWN,0L);
      break;

    case VK_LEFT:
      SendMessage(w,WM_HSCROLL,SB_LINEUP,0L);
      break;

    case VK_RIGHT:
      SendMessage(w,WM_HSCROLL,SB_LINEDOWN,0L);
      break;
    }
  }


/* Local function to set scroll bars up */
static void set_sb(HWND w,MEMWINFO *minfo,UINT wid,
                   UINT hi,int save)
  {
  RECT r;
  if (minfo->flags&V_INIT) return;
  if (minfo->flags&V_AUTOHSCROLL)
    {
    if (minfo->xoff&&minfo->width<=wid)
      {
/* If bitmap will fit in client area, make it do so */
      minfo->xoff=0;
      if (save) save_info(w,minfo);
      }
/* Set up H bar */
    SetScrollPos(w,SB_HORZ,minfo->xoff,FALSE);
    SetScrollRange(w,SB_HORZ,0,
                  (unsigned)
                  max(0L,(long)minfo->width-(long)wid)
                  ,TRUE);
/* Recompute size -- may have changed if scroll bar
   enabled by above step */
    GetClientRect(w,&r);
    wid=r.right-r.left;
    hi=r.bottom-r.top;
    }
  if (minfo->flags&V_AUTOVSCROLL)
    {
    if (minfo->yoff&&minfo->height<=hi)
      {
/* If bitmap will fit in client area, make it do so */
      minfo->yoff=0;
      if (save) save_info(w,minfo);
      }
/* Set up V bar */
    SetScrollPos(w,SB_VERT,minfo->yoff,FALSE);
    SetScrollRange(w,SB_VERT,0,
                (unsigned)max(0L,(long)minfo->height-
                         (long)hi),
                TRUE);
    }
  }




/* Magic paint routine */
static void do_paint(HWND w)
    {
    HDC         hdc;
    PAINTSTRUCT ps;
    MEMWINFO    minfo;
    RECT r;
    int oldmode;
#ifndef WIN32
    DWORD oldworg,oldvorg;
#else
    POINT oldworg,oldvorg;
#endif
    hdc=BeginPaint(w,&ps);
    GetClientRect(w,&r);
    Vget_info(w,&minfo);
    if (minfo.map==NULL||minfo.map->dc==0)
      {
      ReleaseDC(w,hdc);
      EndPaint(w,&ps);
      return;
      }
/* Set up DC the way we like it */
    oldmode=SetMapMode(minfo.map->dc,MM_TEXT);
#ifndef WIN32
    oldworg=SetWindowOrg(minfo.map->dc,0,0);
    oldvorg=SetViewportOrg(minfo.map->dc,0,0);
#else
    SetWindowOrgEx(minfo.map->dc,0,0,&oldworg);
    SetViewportOrgEx(minfo.map->dc,0,0,&oldvorg);
#endif
/* Do something different for scale window */
    if (minfo.flags&V_SCALE)
      {
      int oldmode;
      oldmode=SetStretchBltMode(hdc,minfo.map->stretch_mode);
      StretchBlt(hdc,0,0,r.right-r.left,r.bottom-r.top,
                 minfo.map->dc,minfo.xoff,minfo.yoff,
                 minfo.width,minfo.height,SRCCOPY);
      SetStretchBltMode(hdc,oldmode);
      }
    else
      {
/* if VMAP doesn't entirely cover window, clear first */
      if (r.right-r.left>minfo.width-minfo.xoff||
          r.bottom-r.top>minfo.height-minfo.yoff)
        {
        HBRUSH brush;
        brush=CreateSolidBrush(GetBkColor(minfo.map->dc));
        brush=SelectObject(hdc,brush);
        /* erase "under" bitmap */
        PatBlt(hdc,0,minfo.height,
            r.right-r.left,r.bottom-r.top,PATCOPY);
        /* erase to "right" of bitmap */
        PatBlt(hdc,minfo.width,0,
            r.right-r.left,r.bottom-r.top,PATCOPY);
        DeleteObject(SelectObject(hdc,brush));
        }
/* Draw it */
      BitBlt(hdc,0,0,minfo.width-minfo.xoff,
             minfo.height-minfo.yoff,minfo.map->dc,
             minfo.xoff,minfo.yoff,SRCCOPY);
      }
    SetMapMode(minfo.map->dc,oldmode);
#ifndef WIN32
    SetWindowOrg(minfo.map->dc,LOWORD(oldworg),
                 HIWORD(oldworg));
    SetViewportOrg(minfo.map->dc,LOWORD(oldvorg),
                 HIWORD(oldvorg));
#else
    SetWindowOrgEx(minfo.map->dc,oldworg.x,oldworg.y,NULL);
    SetViewportOrgEx(minfo.map->dc,oldvorg.x,oldvorg.y,NULL);
#endif
    ReleaseDC(w,hdc);
    EndPaint(w,&ps);
    }


